@@ -1,3 +1,45 @@
+"""
+changes.
+
+Usage:
+  changes [options] <app_name> changelog
+  changes [options] <app_name> release
+  changes [options] <app_name> version
+  changes [options] <app_name> test
+  changes [options] <app_name> install
+  changes [options] <app_name> upload
+  changes [options] <app_name> pypi
+  changes [options] <app_name> tag
+
+  changes -h | --help
+
+Options:
+  --new-version=<ver>   Specify version.
+  -p --patch            Patch-level version increment.
+  -m --minor            Minor-level version increment.
+  -M --major            Minor-level version increment.
+
+  -h --help             Show this screen.
+
+  --pypi=<pypi>         Use alternative package index
+  --dry-run             Prints the commands that would have been executed.
+  --skip-changelog      For the release task: should the changelog be generated
+                        and committed?
+  --tox                 Use tox instead of nosetests
+  --test-command=<cmd>  Command to use to test the newly installed package
+  --debug               Debug output.
+
+The commands do the following:
+   changelog   Generates an automatic changelog from your commit messages
+   version     Increments the __version__ attribute of your module's __init__
+   test        Runs your tests with nosetests
+   install     Attempts to install the sdist
+   tag         Tags your git repo with the new version number
+   upload      Uploads your project with setup.py clean sdist upload
+   pypi        Attempts to install your package from pypi
+   release     Runs all the previous commands
+"""
+
 import ast
 import re
 import subprocess
@@ -7,14 +49,15 @@ from docopt import docopt
 from path import path
 import semantic_version
 import logging
+import virtualenv
 
 import changes
 
 
 log = logging.getLogger(__name__)
-
 CHANGELOG = 'CHANGELOG.md'
 
+
 def extract(dictionary, keys):
     """
     Extract only the specified keys from a dict
@@ -52,32 +95,6 @@ def increment(version, major=False, minor=False, patch=True):
     return str(version)
 
 
-def write_new_changelog(app_name, filename, content_lines, dry_run=True):
-    heading_and_newline = (
-        '# [Changelog](%s/releases)\n' %
-        extract_attribute(app_name, '__url__')
-    )
-
-    with open(filename, 'r+') as f:
-        existing = f.readlines()
-
-    output = existing[2:]
-    output.insert(0, '\n')
-
-    for index, line in enumerate(content_lines):
-        output.insert(0, content_lines[len(content_lines) - index - 1])
-
-    output.insert(0, heading_and_newline)
-
-    output = ''.join(output)
-
-    if not dry_run:
-        with open(filename, 'w+') as f:
-            f.write(output)
-    else:
-        log.info('New changelog:\n%s' % output)
-
-
 def get_new_version(app_name, current_version,
                     major=False, minor=False, patch=True):
 
@@ -130,30 +147,41 @@ def current_version(app_name):
 
 
 def execute(commands, dry_run=True):
+    log.debug('executing %s', commands)
     if not dry_run:
         try:
             return subprocess.check_output(commands)
         except subprocess.CalledProcessError, e:
-            return '\n%s' % e.output
+            log.debug('return code: %s, output: %s', e.returncode, e.output)
+            return False
     else:
-        log.debug('execute: %s' % commands)
+        return True
 
 
-def version(arguments):
-    dry_run = arguments['--dry-run']
-    app_name = arguments['<app_name>']
-    new_version = arguments['new_version']
+def write_new_changelog(app_name, filename, content_lines, dry_run=True):
+    heading_and_newline = (
+        '# [Changelog](%s/releases)\n' %
+        extract_attribute(app_name, '__url__')
+    )
 
-    replace_attribute(app_name, '__version__', new_version, dry_run=dry_run)
+    with open(filename, 'r+') as f:
+        existing = f.readlines()
 
-    commands = [
-        'git', 'ci', '-m', new_version,
-        '%s/__init__.py' % app_name, CHANGELOG
-    ]
+    output = existing[2:]
+    output.insert(0, '\n')
 
-    execute(commands, dry_run=dry_run)
+    for index, line in enumerate(content_lines):
+        output.insert(0, content_lines[len(content_lines) - index - 1])
 
-    execute(['git', 'push'], dry_run=dry_run)
+    output.insert(0, heading_and_newline)
+
+    output = ''.join(output)
+
+    if not dry_run:
+        with open(filename, 'w+') as f:
+            f.write(output)
+    else:
+        log.info('New changelog:\n%s' % output)
 
 
 def changelog(arguments):
@@ -192,7 +220,11 @@ def changelog(arguments):
             git_log_content[index] = new_line
 
     if git_log_content:
-        [changelog_content.append('* %s\n' % line) if line else line for line in git_log_content[:-1]]
+        [
+            changelog_content.append('* %s\n' % line)
+            if line else line
+            for line in git_log_content[:-1]
+        ]
 
     write_new_changelog(
         app_name,
@@ -203,12 +235,79 @@ def changelog(arguments):
     log.info('Added content to CHANGELOG.md')
 
 
-def tag(arguments):
+def version(arguments):
     dry_run = arguments['--dry-run']
+    app_name = arguments['<app_name>']
     new_version = arguments['new_version']
 
-    execute(['git', 'tag', '-a', new_version, '-m', '"%s"' % new_version], dry_run=dry_run)
-    execute(['git', 'push', '--tags'], dry_run=dry_run)
+    replace_attribute(
+        app_name,
+        '__version__',
+        new_version,
+        dry_run=dry_run)
+
+
+def commit_version_change(arguments):
+    dry_run = arguments['--dry-run']
+    app_name = arguments['<app_name>']
+    new_version = arguments['new_version']
+
+    commands = [
+        'git', 'ci', '-m', new_version,
+        '%s/__init__.py' % app_name, CHANGELOG
+    ]
+
+    execute(commands, dry_run=dry_run)
+
+    execute(['git', 'push'], dry_run=dry_run)
+
+
+def test(arguments):
+    dry_run = arguments['--dry-run']
+    command = 'nosetests'
+    if arguments['--tox']:
+        command = 'tox'
+
+    return execute([command], dry_run=dry_run)
+
+
+def make_virtualenv():
+    tmp_dir = tempfile.mkdtemp()
+    log.debug('tmp ve dir: %s' % tmp_dir)
+    virtualenv.create_environment(tmp_dir, site_packages=False)
+    return tmp_dir
+
+
+def run_test_command(arguments):
+    if arguments['--test-command']:
+        test_command = arguments['--test-command'].split(' ')
+        result = execute(test_command, dry_run=arguments['--dry-run'])
+        log.info('Test command "%s" result: %s', test_command, result)
+
+
+def install(arguments):
+    dry_run = arguments['--dry-run']
+    app_name = arguments['<app_name>']
+    new_version = arguments['new_version']
+
+    result = execute(
+        ['python', 'setup.py', 'clean', 'sdist'],
+        dry_run=dry_run
+    )
+    if result:
+        tmp_dir = make_virtualenv()
+        try:
+            virtualenv.install_sdist(
+                arguments['<app_name>'],
+                'dist/%s-%s.tar.gz' % (app_name, new_version),
+                '%s/bin/python' % tmp_dir
+            )
+            log.info('Successfully installed %s sdist', app_name)
+            run_test_command(arguments)
+        except:
+            log.info('Error installing %s sdist', app_name)
+
+        path(tmp_dir).rmtree(path(tmp_dir))
 
 
 def upload(arguments):
@@ -216,68 +315,97 @@ def upload(arguments):
     pypi = arguments['--pypi']
 
     upload = ['python', 'setup.py', 'clean', 'sdist', 'upload']
-    if pypi: 
+    if pypi:
         upload.append('-r')
         upload.append(pypi)
 
     execute(upload, dry_run=dry_run)
 
 
+def pypi(arguments):
+    dry_run = arguments['--dry-run']
+    app_name = arguments['<app_name>']
+    pypi = arguments['--pypi']
+    package_index = 'pypi'
+
+    tmp_dir = make_virtualenv()
+
+    install = ['%s/bin/pip' % tmp_dir, 'install', app_name]
+    if pypi:
+        install.append('-i')
+        install.append(pypi)
+        package_index = pypi
+
+    try:
+        result = execute(install, dry_run=dry_run)
+        if result:
+            log.info('Successfully installed %s from %s',
+                     app_name, package_index)
+        else:
+            log.error('Failed to install %s from %s',
+                      app_name, package_index)
+        run_test_command(arguments)
+    except:
+        log.exception('')
+        log.info('Error installing %s from %s', app_name, package_index)
+
+    path(tmp_dir).rmtree(path(tmp_dir))
+
+
+def tag(arguments):
+    dry_run = arguments['--dry-run']
+    new_version = arguments['new_version']
+
+    execute(
+        ['git', 'tag', '-a', new_version, '-m', '"%s"' % new_version],
+        dry_run=dry_run
+    )
+    execute(['git', 'push', '--tags'], dry_run=dry_run)
+
+
 def release(arguments):
-    if not arguments['--skip-changelog']:
-        changelog(arguments)
-    version(arguments)
-    tag(arguments)
-    upload(arguments)
-
-
-cli = """
-changes.
-
-Usage:
-  changes [options] <app_name> changelog
-  changes [options] <app_name> release
-  changes [options] <app_name> version
-  changes [options] <app_name> tag
-  changes [options] <app_name> upload
-
-  changes -h | --help
-
-Options:
-  --new-version=<ver>   Specify version.
-  -p --patch            Patch-level version increment.
-  -m --minor            Minor-level version increment.
-  -M --major            Minor-level version increment.
-
-  -h --help             Show this screen.
-
-  --pypi=<pypi>         Specify alternative pypi
-  --dry-run             Prints the commands that would have been executed.
-  --skip-changelog      For the release task: should the changelog be generated
-                        and committed?
-  --debug               Debug output.
-"""
+    try:
+        if not arguments['--skip-changelog']:
+            changelog(arguments)
+        version(arguments)
+        test(arguments)
+        commit_version_change(arguments)
+        install(arguments)
+        upload(arguments)
+        pypi(arguments)
+        tag(arguments)
+    except:
+        log.error('Error releasing')
 
 
 def main():
-    arguments = docopt(cli, version=changes.__version__)
+    commands = ['release', 'changelog', 'test', 'version', 'tag', 'upload',
+                'install', 'pypi']
+
+    arguments = docopt(__doc__, version=changes.__version__)
     debug = arguments['--debug']
     logging.basicConfig(level=logging.DEBUG if debug else logging.INFO)
 
+    suppress_version_prompt_for = ['test', 'upload']
+
     app_name = arguments['<app_name>']
+
     if arguments['--new-version']:
-        new_version = arguments['--new-version']
-    else:
-        new_version = get_new_version(
-            app_name,
-            current_version(app_name),
-            **dict([
-                (key[2:], value) for key, value in extract(arguments, ['--major', '--minor', '--patch']).items()
-            ])
-        )
+        arguments['new_version'] = arguments['--new-version']
 
-    arguments['new_version'] = new_version
     log.debug('arguments: %s', arguments)
-    for command in ['release', 'version', 'changelog', 'tag', 'upload']:
+
+    for command in commands:
         if arguments[command]:
+            if command not in suppress_version_prompt_for:
+                arguments['new_version'] = get_new_version(
+                    app_name,
+                    current_version(app_name),
+                    **dict([
+                        (key[2:], value)
+                        for key, value in
+                        extract(arguments, ['--major', '--minor', '--patch'])
+                        .items()
+                    ])
+                )
             globals()[command](arguments)
